import asyncio
import importlib.util
import pathlib
import sys
import traceback
from types import ModuleType
from typing import Literal, Sequence

import fire
from fire.core import FireExit
from rich.console import Console
from rich.style import Style
from rich.text import Text

from glhf.client import BasicClient
from glhf.base import ClientProtocol, ServerProtocol
from glhf.gui import PygameGUI
from glhf.server import LocalServer, SocketioServer


class CLI:
    """Command Line Interface for managing clients and starting the server"""

    def __init__(self, module: ModuleType, console: Console) -> None:
        self._module = module
        self._console = console
        self.__lazy_init__()

    def __lazy_init__(self) -> None:
        self._server: ServerProtocol | None = None
        self._clients: list[ClientProtocol] = []

    def show(self) -> None:
        """Display the server and client information."""
        self._console.print(self._server, self._clients)

    def start(self, debug: bool = False) -> None:
        """Start the server.

        Args:
            debug (bool, optional): Whether to enable debug mode. Defaults to False.

        """
        if self._server is None:
            raise RuntimeError("server not set")
        asyncio.run(start(self._server, self._clients), debug=debug)
        self.__lazy_init__()

    def server(self, name: str | Literal["local", "socketio"] = "local") -> None:
        """Set the server type.

        Args:
            name (str | Literal["local", "socketio"], optional): The server type to set. Defaults to "socketio".

        Raises:
            ValueError: If the provided server name is unknown.

        """

        if name == "local":
            self._server = LocalServer(15, 15)
        elif name == "socketio":
            self._server = SocketioServer()
        else:
            try:
                cls = getattr(self._module, name)
            except AttributeError:
                raise ValueError(f"unknown server: {name}")
            else:
                self._server = cls()

    def client(self, bot: str) -> None:
        """Add a bot to the client list.

        Args:
            name (str): The name of the bot to add.

        Raises:
            ValueError: If the provided bot name is unknown.

        """

        if self._server is None:
            raise RuntimeError("server not set")

        try:
            bot_cls = getattr(self._module, bot)
        except AttributeError:
            raise ValueError(f"unknown bot: {bot}")
        else:
            bot_obj = bot_cls()
            USERID = "h4K1gOyHNnkGngym8fUuYA"
            USERNAME = "PsittaTestBot"
            gui = PygameGUI()
            client = BasicClient(USERID, USERNAME, bot_obj, gui, self._server)
            self._clients.append(client)


def set_eager_task_factory(is_eager: bool) -> None:
    loop = asyncio.get_event_loop()
    loop.set_task_factory(asyncio.eager_task_factory if is_eager else None)  # type: ignore


async def start(server: ServerProtocol, clients: Sequence[ClientProtocol]) -> None:
    set_eager_task_factory(True)

    async with asyncio.TaskGroup() as g:
        for client in clients:
            g.create_task(client.run())


def main(file: str = "") -> None:

    if file:
        location = pathlib.Path(file)
        name = location.stem
        spec = importlib.util.spec_from_file_location(name, location)
        if spec is None:
            raise IOError(f"could not load module from {location}")
        module = importlib.util.module_from_spec(spec)
        sys.modules[name] = module
        assert spec.loader is not None, "spec.loader is None"
        spec.loader.exec_module(module)
    else:
        # module = importlib.import_module('__main__')
        import __main__ as module

    console = Console()
    cli = CLI(module, console)

    name = "GLHF"
    prompt = Text.assemble(
        (name, Style(color="#008080", bold=True)),
        ("> ", Style(color="white", bold=True)),
    )

    while True:
        try:
            fire.Fire(cli, console.input(prompt), name)

        except FireExit:
            pass

        except EOFError:
            console.print("GG!", style="b red")
            break

        except Exception:
            traceback.print_exc()

        finally:
            console.print()


if __name__ == "__main__":
    main()



__all__ = (
    "patch",
    "make_diff",
)


def coord_to_name(coord: tuple[int, int]) -> str:
    """Converts a coordinate to a string representation."""
    return f"{coord[0]},{coord[1]}"


def coord_to_index(coord: tuple[int, int], col: int) -> int:
    """Converts a coordinate to an index in a 1D list representation."""
    return coord[0] * col + coord[1]


def patch(old: list[int], diff: list[int]) -> list[int]:
    new = []
    i = 0
    len_diff = len(diff)
    while i < len_diff:
        n = diff[i]
        if n:
            j = len(new)
            new += old[j : j + n]
        i += 1
        if i == len_diff:
            break
        n = diff[i]
        if n:
            j = i + 1
            new += diff[j : j + n]
            i += n
        i += 1
    return new


def make_diff(new: list[int], old: list[int]) -> list[int]:
    diff = []
    i = 0
    j = 0
    len_new = len(new)
    len_old = len(old)
    len_min = min(len_new, len_old)
    while True:
        while j < len_min and new[j] == old[j]:
            j += 1
        diff.append(j - i)
        i = j
        if j == len_min:
            if len_new > len_old:
                diff.append(len_new - len_old)
                diff += new[len_old:]
            break
        while j < len_min and new[j] != old[j]:
            j += 1
        if j < len_min:
            diff.append(j - i)
            diff += new[i:j]
            i = j
        else:
            diff.append(len_new - i)
            diff += new[i:len_new]
            break
    return diff


class Map(list[int]):
    @property
    def shape(self) -> list[int]:
        return self[:2:-1]

    @property
    def armies(self) -> list[list[int]]:
        m = self
        c, r = m[:2]
        stop = 2 + c * r
        return [m[i : i + r] for i in range(2, stop, r)]

    @property
    def terrain(self) -> list[list[int]]:
        m = self
        c, r = m[:2]
        n = c * r
        start = 2 + n
        return [m[i : i + r] for i in range(start, start + n, r)]


class Cities(list[int]):
    @property
    def cities(self) -> list[int]:
        return self.copy()



from functools import lru_cache
from itertools import product
from os import PathLike
from pathlib import Path
from threading import Event, Thread
from typing import IO, Sequence

import pygame
from pygame import (
    Color,
    Rect,
    Surface,
    Vector2,
    display,
    draw,
    event,
    font,
    image,
    mouse,
    sprite,
    time,
    transform,
)

from glhf.helper import patch
from glhf.typing_ import GameStartDict, GameUpdateDict
from glhf.utils import signalize, streamify

__all__ = ("PygameGUI",)

type AnyPath = str | bytes | PathLike[str] | PathLike[bytes]
type FileArg = AnyPath | IO[bytes] | IO[str]
type Coordinate = Sequence[float]

PALETTE = (
    0x222222FF,
    0x393939FF,
    0x808080FF,
    0xBBBBBBFF,
    0xDCDCDCFF,
    0xFF0000FF,
    0x4363D8FF,
    0x008000FF,
    0x008080FF,
    0xF58231FF,
    0xF032E6FF,
    0x800080FF,
    0x800000FF,
    0xB09F30FF,
    0x9A6324FF,
    0x0000FFFF,
    0x483D8BFF,
)
PALETTE = tuple(map(Color, PALETTE))
ARROWS = "↑↓←→"
WHITE = Color("white")
BLACK = Color("black")

path = Path(__file__).parent


class CachedResource:
    __slots__ = "font_", "city", "crown", "mountain", "obstacle"

    def __init__(
        self,
        font_: FileArg = path / "resource/Quicksand-Bold.ttf",
        city: FileArg = path / "resource/city.png",
        crown: FileArg = path / "resource/crown.png",
        mountain: FileArg = path / "resource/mountain.png",
        obstacle: FileArg = path / "resource/obstacle.png",
    ) -> None:
        self.font_ = font_
        load = image.load
        self.city = load(city).convert_alpha()
        self.crown = load(crown).convert_alpha()
        self.mountain = load(mountain).convert_alpha()
        self.obstacle = load(obstacle).convert_alpha()

    def render_cell(self, size: int, terrain: int, army: int) -> Surface:
        """
        Render the terrain without army.

        Args:
            size (int): The size of the cell.
            terrain (int): The terrain value.

        Returns:
            Surface: The rendered cell.

        Raises:
            ValueError: If the terrain value is not within the valid range.

        Notes:
            The meaning of different terrain values are as follows:
            -4: fog obstacle
            -3: fog
            -2: mountain
            -1: neutral without army
            0 ~ 11: lands
            12 ~ 23: cities
            24 ~ 35: crowns
            36: neutral with army

        """
        if army and terrain >= 0:
            return self.render_terrain_army(size, terrain, army)
        return self.render_terrain(size, terrain)

    @lru_cache(256)
    def render_terrain_army(self, size: int, terrain: int, army: int) -> Surface:
        cell = self.render_terrain(size, terrain).copy()
        cell.blit(*self.render_army(size, army))
        return cell

    @lru_cache(16)
    def get_font(self, size: int) -> font.Font:
        return font.Font(self.font_, round((size - 12) ** 0.5) + 10)

    @lru_cache(512)
    def render_army(self, size: int, army: int) -> tuple[Surface, Rect]:
        s = str(army)
        render = self.get_font(size).render
        text = render(s, True, WHITE)
        shadow = render(s, True, BLACK)
        surf = transform.box_blur(shadow, 2)
        surf.blit(text, (0, 0))
        m = size // 2
        rect = surf.get_rect(center=(m, m))
        return surf, rect

    @lru_cache(256)
    def render_terrain(self, size: int, terrain: int) -> Surface:
        if terrain < 0:
            if terrain < -2:
                cell = self.render_index(size, 1)
                if terrain == -4:
                    cell.blit(*self.render_image(size, self.obstacle))
            elif terrain < -1:
                cell = self.render_index(size, 3)
                cell.blit(*self.render_image(size, self.mountain))
            else:
                cell = self.render_index(size, 4)
        elif terrain < 36:
            q, r = divmod(terrain, 12)
            cell = self.render_index(size, r + 5)
            if q == 1:
                cell.blit(*self.render_image(size, self.city))
            elif q == 2:
                cell.blit(*self.render_image(size, self.crown))
        else:
            cell = self.render_index(size, 2)
            if terrain == 37:
                cell.blit(*self.render_image(size, self.city))
        return cell

    @classmethod
    def render_index(cls, size: int, index: int) -> Surface:
        cell = Surface((size, size))
        if index == 1:
            cell.fill(PALETTE[1])
        else:
            s = size - 1
            draw.rect(cell, PALETTE[index], (1, 1, s, s))
            draw.rect(cell, BLACK, (0, 0, size, size), 1)
        return cell

    @classmethod
    def render_image(cls, size: int, image: Surface) -> tuple[Surface, Rect]:
        s = size * 25 // 32
        m = size // 2
        surf = transform.smoothscale(image, (s, s))
        rect = surf.get_rect(center=(m, m))
        return surf, rect

    @classmethod
    def all_cache_info(cls):
        return {
            "font": cls.get_font.cache_info(),
            "army": cls.render_army.cache_info(),
            "terrain": cls.render_terrain.cache_info(),
            "all": cls.render_terrain_army.cache_info(),
        }


class BackgroundSprite(sprite.DirtySprite):
    def __init__(self, *group) -> None:
        super().__init__(*group)
        self.image = Surface((0, 0))
        self.rect = self.image.get_rect()
        self.layer = 0

    def fast_update(
        self,
        window_resize: bool,
        window_size: tuple[int, int] | None = None,
    ) -> None:
        assert isinstance(self.rect, Rect)
        if window_resize:
            window_size = window_size or display.get_window_size()
            self.image = Surface(window_size)
            self.rect.size = window_size
            self.image.fill(PALETTE[0])
            self.dirty = 1


class MapSprite(sprite.DirtySprite):
    # image: Surface
    # rect: Rect

    def __init__(
        self,
        *groups,
        window_size: tuple[int, int] | None = None,
    ) -> None:
        super().__init__(*groups)
        x, y = window_size or display.get_window_size()
        self.zoom = 3
        self.cell_size = 12
        self.center = Vector2(x // 2, y // 2)
        self.image = Surface((0, 0))
        self.rect = self.image.get_rect(center=self.center)
        self.data = None
        self.layer = 1
        self.resource = CachedResource()

        self.map_: list[int] = [0, 0]
        self.cities: list[int] = []

        # self._turn = 0

    @lru_cache(16)
    def get_image(self, size: tuple[int, int]) -> Surface:
        return Surface(size)

    def fast_update(
        self,
        zoom: int,
        move: tuple[float, float] | Vector2,
        data: GameUpdateDict | None,
    ) -> None:
        assert isinstance(self.rect, Rect)
        assert isinstance(self.image, Surface)

        skip = True
        resize = False

        if data:
            # if self._turn + 1 < data.turn:
            #     print(f"missing turn {self._turn + 1} to {data.turn - 1}")
            # self._turn = data.turn

            shape = self.map_[:2]
            self.map_ = patch(self.map_, data["map_diff"])
            self.cities = patch(self.cities, data["cities_diff"])
            if self.map_[:2] != shape:
                resize = True

            skip = False
            self.data = data

        if not self.data:
            return

        if zoom and self.zoom != (z := max(0, min(self.zoom + zoom, 15))):
            skip = False
            resize = True
            s = 12 + z * z
            p = mouse.get_pos()
            c = self.center
            c -= p
            c *= s / self.cell_size
            c += p
            self.zoom = z
            self.cell_size = s

        if move and move != (0, 0):
            skip = False
            self.center += move
            self.rect.center = self.center

        if skip:
            return

        if resize:
            data = self.data
            c, r = self.map_[:2]
            c_size = self.cell_size
            size = c * c_size, r * c_size
            rect = self.rect
            rect.size = size
            rect.center = self.center
            self.image = self.get_image(size)

        if data:
            img = self.image
            res = self.resource
            c_size = self.cell_size

            c, r = self.rect.size
            yx = tuple(product(range(0, r, c_size), range(0, c, c_size)))

            c, r = self.map_[:2]
            n = c * r
            armies = self.map_[2 : 2 + n]
            terrain = self.map_[2 + n :]

            for i in self.cities:
                if terrain[i] >= 0:
                    terrain[i] += 12
                elif armies[i]:
                    terrain[i] = 37

            for i in data["generals"]:
                if i >= 0 and terrain[i] >= 0:
                    terrain[i] += 24

            rr = self.rect.clip((0, 0), display.get_window_size())
            x, y = self.rect.topleft
            rr.move_ip(-x, -y)

            left = rr.left - self.cell_size
            right = rr.right
            top = rr.top - self.cell_size
            bottom = rr.bottom

            for (y, x), t, a in zip(yx, terrain, armies):
                if not (left <= x < right and top <= y < bottom):
                    continue
                if t == -1 and a:
                    t = 36
                surf = res.render_cell(c_size, t, a)
                img.blit(surf, (x, y))

        self.dirty = 1


class PygameGUI(Thread):
    WINDOWSIZE = 800, 600

    @streamify
    def game_start(self, data: GameStartDict) -> GameStartDict:
        return data

    @streamify
    def game_update(self, data: GameUpdateDict) -> GameUpdateDict:
        return data

    @signalize
    def game_over(self) -> None:
        self.game_update.close()

    def __init__(self, **kwargs) -> None:
        super().__init__(name=f"{type(self).__name__}-{id(self):X}", **kwargs)
        self.quit = Event()

    def __enter__(self) -> None:
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if exc_type is None:
            self.quit.wait()
        self.join()

    def main(self, window: Surface) -> None:
        clock = time.Clock()

        bg = BackgroundSprite()
        bg.fast_update(True)
        map_ = MapSprite()

        group = sprite.LayeredDirty(bg, map_)
        rects = group.draw(window)
        display.update(rects)

        move = Vector2()

        while not self.quit.is_set():
            window_resize = False
            zoom = 0
            move.update()

            for e in event.get():
                if e.type == pygame.QUIT:
                    self.quit.set()
                elif e.type == pygame.WINDOWSIZECHANGED:
                    window_resize = True
                elif e.type == pygame.MOUSEWHEEL:
                    zoom += e.x + e.y
                elif e.type == pygame.MOUSEMOTION:
                    if e.buttons[0] == 1:
                        move += e.rel

            bg.fast_update(window_resize)
            map_.fast_update(zoom, move, self.game_update.get())

            rects = group.draw(window)
            if rects:
                display.update(rects)

            clock.tick(60)
            # print(f"\r{clock.get_fps():0>7.1f}", end="")

    def run(self) -> None:
        pygame.init()
        # pygame.RESIZABLE | pygame.SCALED
        try:
            window = display.set_mode(self.WINDOWSIZE)
            while not self.quit.is_set():
                for e in event.get():
                    if e.type == pygame.QUIT:
                        return
                if self.game_start.get() is not None:
                    self.main(window)
        finally:
            pygame.quit()
            self.quit.set()


if __name__ == "__main__":

    def main() -> None:
        pass

    main()



from __future__ import annotations

import asyncio
from abc import abstractmethod
from typing import Any, Protocol

from glhf.typing_ import GameStartDict, GameUpdateDict, QueueUpdateDict
from glhf.utils import methodlike

__all__ = "ServerProtocol", "ClientProtocol", "BotProtocol"


class ServerProtocol(Protocol):
    # ============================================================
    # send
    # ============================================================

    @abstractmethod
    def set_username(
        self,
        client: ClientProtocol,
        user_id: str,
        username: str,
    ) -> asyncio.Task[None]: ...

    @abstractmethod
    def stars_and_rank(
        self,
        client: ClientProtocol,
        user_id: str,
    ) -> asyncio.Task[None]: ...

    @abstractmethod
    def join_private(
        self,
        client: ClientProtocol,
        queue_id: str,
        user_id: str,
    ) -> asyncio.Task[None]: ...

    @abstractmethod
    def set_force_start(
        self,
        client: ClientProtocol,
        queue_id: str,
        do_force: bool,
    ) -> asyncio.Task[None]: ...

    @abstractmethod
    def leave_game(
        self,
        client: ClientProtocol,
    ) -> asyncio.Task[None]: ...

    @abstractmethod
    def surrender(
        self,
        client: ClientProtocol,
    ) -> asyncio.Task[None]: ...

    @abstractmethod
    def attack(
        self, client: ClientProtocol, start: int, end: int, is50: bool
    ) -> asyncio.Task[None]: ...

    # ============================================================
    # run
    # ============================================================

    @abstractmethod
    async def connect(
        self,
        client: ClientProtocol,
        user_id: str,
    ) -> None: ...

    @abstractmethod
    async def disconnect(
        self,
        client: ClientProtocol,
        user_id: str,
    ) -> None: ...


class ClientProtocol(Protocol):
    # ============================================================
    # recieve
    # ============================================================

    @abstractmethod
    def stars(self, data: dict[str, float]) -> Any: ...

    @abstractmethod
    def rank(self, data: dict[str, int]) -> Any: ...

    @abstractmethod
    def chat_message(self, chat_room: str, data: dict[str, Any]) -> Any: ...

    @abstractmethod
    def notify(self, data: Any, _: Any = None) -> Any: ...

    @abstractmethod
    def queue_update(self, data: QueueUpdateDict) -> Any: ...

    @abstractmethod
    def pre_game_start(self) -> Any: ...

    @abstractmethod
    def game_start(self, data: GameStartDict, _: Any = None) -> Any: ...

    @abstractmethod
    def game_update(self, data: GameUpdateDict, _: Any = None) -> Any: ...

    @abstractmethod
    def game_won(self, _1: Any = None, _2: Any = None) -> Any: ...

    @abstractmethod
    def game_lost(self, _1: Any = None, _2: Any = None) -> Any: ...

    @abstractmethod
    def game_over(self, _1: Any = None, _2: Any = None) -> Any: ...

    # ============================================================
    # send
    # ============================================================

    @abstractmethod
    def set_username(self) -> asyncio.Task[None]: ...

    @abstractmethod
    def stars_and_rank(self) -> asyncio.Task[None]: ...

    @abstractmethod
    def join_private(self, queue_id: str) -> asyncio.Task[None]: ...

    @abstractmethod
    def set_force_start(self, do_force: bool) -> asyncio.Task[None]: ...

    @abstractmethod
    def leave_game(self) -> asyncio.Task[None]: ...

    @abstractmethod
    def surrender(self) -> asyncio.Task[None]: ...

    @abstractmethod
    def attack(self, start: int, end: int, is50: bool) -> asyncio.Task[None]: ...

    # ============================================================
    # run
    # ============================================================

    @abstractmethod
    async def run(self) -> None: ...


class BotProtocol(Protocol):
    # ============================================================
    # recieve
    # ============================================================

    @methodlike
    @abstractmethod
    def stars(self, data: dict[str, float]) -> Any: ...

    @methodlike
    @abstractmethod
    def rank(self, data: dict[str, int]) -> Any: ...

    @methodlike
    @abstractmethod
    def chat_message(self, chat_room: str, data: dict[str, Any]) -> Any: ...

    @methodlike
    @abstractmethod
    def notify(self, data: Any) -> Any: ...

    @methodlike
    @abstractmethod
    def queue_update(self, data: QueueUpdateDict) -> Any: ...

    @methodlike
    @abstractmethod
    def pre_game_start(self) -> Any: ...

    @methodlike
    @abstractmethod
    def game_start(self, data: GameStartDict) -> Any: ...

    @methodlike
    @abstractmethod
    def game_update(self, data: GameUpdateDict) -> Any: ...

    @methodlike
    @abstractmethod
    def game_won(self) -> Any: ...

    @methodlike
    @abstractmethod
    def game_lost(self) -> Any: ...

    @methodlike
    @abstractmethod
    def game_over(self) -> Any: ...

    # ============================================================
    # run
    # ============================================================

    @abstractmethod
    async def run(self, client: ClientProtocol) -> None: ...


class GUIProtocol(Protocol): ...



from typing import Any, Literal, NamedTuple, TypedDict

__all__ = "QueueUpdateDict", "GameStartDict", "GameUpdateDict", "GameUpdateTuple"


class QueueUpdateDict(TypedDict):
    playerIndices: list[int]
    playerColors: list[int]
    lobbyIndex: int
    isForcing: bool
    numPlayers: int
    numForce: list[int]
    teams: list[int]
    usernames: list[None] | list[str]
    options: dict[str, Any]


class GameStartDict(TypedDict):
    playerIndex: int
    playerColors: list[int]
    replay_id: str
    chat_room: str
    usernames: list[str]
    teams: list[int]
    game_type: Literal["ffa", "1v1", "custom"]
    swamps: list[Any]
    lights: list[Any]
    options: dict[str, Any]


class GameUpdateDict(TypedDict):
    scores: list[dict]
    turn: int
    stars: list[int]
    attackIndex: int
    generals: list[int]
    map_diff: list[int]
    cities_diff: list[int]


class GameUpdateTuple(NamedTuple):
    turn: int
    map_: list[int]
    cities: list[int]
    generals: list[int]



import asyncio
import time
from collections import deque
from dataclasses import dataclass, field
from itertools import combinations as C
from itertools import product as P
from random import Random
from typing import Any

import igraph as ig
import ortools.sat.python.cp_model as cp
from bidict import KeyDuplicationError, ValueDuplicationError, bidict
from socketio import AsyncClient

from glhf.base import ClientProtocol, ServerProtocol
from glhf.helper import make_diff
from glhf.utils import to_task

WSURL = "wss://ws.generals.io"
BOTKEY = "sd09fjd203i0ejwi_changeme"


def make_2d_grid(
    row: int,
    col: int,
) -> ig.Graph:
    """Creates a 2D grid graph.

    Args:
        row: The number of rows in the map.
        col: The number of columns in the map.

    Returns:
        graph: The 2D grid graph.
    """
    g = ig.Graph.Lattice((col, row), circular=False)
    coords = [(x, y) for y, x in P(range(row), range(col))]
    vs = g.vs
    vs["coord"] = coords
    vs["name"] = [f"{x},{y}" for x, y in coords]
    vs["army"] = 0
    vs["terrain"] = -1
    # vs["is_city"] = False
    # vs["is_general"] = False
    return g


def make_complete_map(
    row: int,
    col: int,
) -> tuple[list[int], ig.Graph]:
    """Creates a complete map and its corresponding graph.

    Args:
        row: The number of rows in the map.
        col: The number of columns in the map.

    Returns:
        A tuple containing the map and the graph.
    """
    g = ig.Graph.Lattice((col, row), circular=False)
    g["row"] = row
    g["col"] = col
    coords = [(x, y) for y, x in P(range(row), range(col))]
    vs = g.vs
    vs["coord"] = coords
    vs["name"] = [f"{x},{y}" for x, y in coords]
    n = col * row
    m = [col, row, *(0,) * n, *(-1,) * n]
    return m, g


def p_dispersion[T](
    p: int,
    d_min: int,
    d_max: int,
    selects: list[T],
    coords: list[tuple[int, int]],
    verbose: bool,
) -> list[T]:
    """Finds a dispersion of points on a grid.

    Args:
        p: The number of points to select.
        d_min: The minimum distance between any two selected points.
        d_max: The maximum distance between any two selected points.
        selects: The list of selectable items.
        coords: The coordinates of the selectable items.
        callback: An optional callback function for the solver.

    Returns:
        A list of selected items.

    Raises:
        ValueError: If the solver fails to find a solution.
    """
    model = cp.CpModel()
    new_int_var = model.new_int_var
    new_bool_var = model.new_bool_var
    add = model.add
    Sum = cp.LinearExpr.Sum

    d_min_var = new_int_var(d_min, d_max, "d_min")
    select_vars = [new_bool_var(str(c)) for c in coords]
    add(Sum(select_vars) == p)
    for (i, j), ((x1, y1), (x2, y2)) in zip(C(select_vars, 2), C(coords, 2)):
        d = abs(x1 - x2) + abs(y1 - y2)
        add((2 - i - j) * d_min + d >= d_min_var)  # type: ignore

    solver = cp.CpSolver()
    callback = cp.VarArraySolutionPrinter((d_min_var,)) if verbose else None

    if solver.solve(model, callback) == cp.OPTIMAL:
        boolean_value = solver.boolean_value
        return [i for i, v in zip(selects, select_vars) if boolean_value(v)]

    raise ValueError(solver.status_name())


def move_army(
    graph: ig.Graph,
    player_id: int,
    start: int,
    end: int,
    is50: bool,
) -> bool:
    if graph.are_connected(start, end):
        vs = graph.vs
        u = vs[start]
        v = vs[end]
        u_army = u["army"]
        v_army = v["army"]
        u_terrain = u["terrain"]
        v_terrain = v["terrain"]
        if u_army > 1 and u_terrain == player_id and v_terrain >= -1:
            if is50:
                army = u_army // 2
                u["army"] = army
                army = u_army - army
            else:
                u["army"] = 1
                army = u_army - 1
            if u_terrain == v_terrain:
                v["army"] += army
            else:
                if v_army >= army:
                    v["army"] -= army
                else:
                    v["army"] = army - v_army
                    v["terrain"] = u_terrain
            return True
    return False


@dataclass(slots=True)
class Queue:
    id: str = ""
    force_start: bool = False

    def clear(self) -> None:
        self.id = ""
        self.force_start = False


@dataclass(slots=True)
class Game:
    id: str = ""
    left: bool = False
    surrendered: bool = False
    attacks: deque[tuple[int, int, bool]] = field(default_factory=deque)


@dataclass(slots=True)
class User:
    id: str
    name: str = "Anonymous"


@dataclass(slots=True)
class Player:
    client: ClientProtocol
    user: User
    queue: Queue = field(default_factory=Queue)
    game: Game = field(default_factory=Game)


class LocalServer(ServerProtocol):
    def __init__(
        self,
        row: int,
        col: int,
        moves_per_turn: int = 2,
        turns_per_round: int = 25,
        turns_per_sec: float = 1.0,
    ) -> None:
        self.row = row
        self.col = col

        self.user_ids: bidict[ClientProtocol, str] = bidict()
        self.clients: bidict[str, ClientProtocol] = self.user_ids.inverse

        self.users: dict[str, User] = {}
        self.players: dict[str, Player] = {}
        self.game_queues: dict[str, list[Player]] = {}

        # settings
        self.mpt = moves_per_turn
        self.tpr = turns_per_round
        self.tps = turns_per_sec

    def random_connected_map(
        self,
        graph: ig.Graph,
        seed: Any = None,
    ) -> None:
        es = graph.es
        vs = graph.vs
        rng = Random(seed)
        es["weight"] = [rng.random() for _ in range(graph.ecount())]
        mst = graph.spanning_tree("weight")
        leaves = mst.vs.select(_degree_le=1).indices
        graph.delete_edges(_source=leaves)
        vs[leaves]["terrain"] = -2

    def set_generals(self, graph: ig.Graph, generals: list[int]) -> None:
        vs = graph.vs
        for i, v in enumerate(vs[generals]):
            v["terrain"] = i

    def dispersion_generals(
        self,
        n: int,
        row: int,
        col: int,
        graph: ig.Graph,
        *,
        d_min: int = 9,
        seed: Any = None,
        verbose: bool = False,
    ) -> list[int]:
        vs = graph.vs

        # sample vertices
        rng = Random(seed)
        indices = vs.select(_degree_gt=0).indices
        k = min(max(24, n * 12), len(indices))
        selects = rng.sample(indices, k)

        # update generals
        d_max = row + col - 2
        coords = vs[selects]["coord"]
        generals = p_dispersion(n, d_min, d_max, selects, coords, verbose)
        self.set_generals(graph, generals)
        return generals

    # ============================================================
    # send
    # ============================================================

    @to_task
    async def stars_and_rank(self, client: ClientProtocol, user_id: str) -> None:
        pass

    @to_task
    async def set_username(
        self, client: ClientProtocol, user_id: str, username: str
    ) -> None:
        if self.user_ids[client] != user_id:
            raise
        self.users[user_id].name = username

    @to_task
    async def join_private(
        self, client: ClientProtocol, queue_id: str, user_id: str
    ) -> None:
        if self.user_ids[client] != user_id:
            raise
        game_queue = self.game_queues.setdefault(queue_id, [])
        player = self.players[user_id]
        player.queue.id = queue_id
        game_queue.append(player)
        self._queue_update(queue_id)

    @to_task
    async def set_force_start(
        self, client: ClientProtocol, queue_id: str, do_force: bool
    ) -> None:
        try:
            user_id = self.user_ids[client]
        except KeyError:
            raise
        players = self.game_queues[queue_id]
        for player in players:
            if player.user.id == user_id:
                player.queue.force_start = do_force
                break
        else:
            raise
        self._queue_update(queue_id)

    @to_task
    async def leave_game(self, client: ClientProtocol) -> None:
        try:
            user_id = self.user_ids[client]
        except KeyError:
            raise
        player = self.players[user_id]
        if player.game.id:
            player.game.left = True
        else:
            raise

    @to_task
    async def surrender(self, client: ClientProtocol) -> None:
        try:
            user_id = self.user_ids[client]
        except KeyError:
            raise
        player = self.players[user_id]
        if player.game.id:
            player.game.surrendered = True
        else:
            raise

    @to_task
    async def attack(
        self, client: ClientProtocol, start: int, end: int, is50: bool
    ) -> None:
        try:
            user_id = self.user_ids[client]
        except KeyError:
            raise
        player = self.players[user_id]
        player.game.attacks.append((start, end, is50))

    # ============================================================
    # recieve
    # ============================================================

    def _queue_update(self, queue_id: str) -> None:
        try:
            players = self.game_queues[queue_id]
        except KeyError:
            raise

        for i, player in enumerate(players):
            data = {
                "playerIndices": i,
                "isForcing": player.queue.force_start,
                "usernames": [player.user.name for player in players],
            }
            player.client.queue_update(data)  # type: ignore

        if sum(user_data.queue.force_start for user_data in players) >= 1:
            del self.game_queues[queue_id]
            for player in players:
                player.queue.clear()
                player.game.id = queue_id
            asyncio.create_task(self.run(players))

    def _game_start(self, players: list[Player]) -> None:
        for i, player in enumerate(players):
            player.client.game_start({"playerIndex": i})  # type: ignore

    def _game_update(self, players: list[Player], turn, generals, map_diff) -> None:
        for player in players:
            player.client.game_update(
                {
                    "scores": [],
                    "turn": turn,
                    "stars": [],
                    "attackIndex": 0,
                    "generals": generals,
                    "map_diff": map_diff,
                    "cities_diff": [],
                }
            )

    def _game_over(self, players: list[Player]) -> None:
        for player in players:
            player.client.game_over()

    def _map_update(
        self,
        players: list[Player],
        row: int,
        col: int,
        turn: int,
        map_old: list[int],
        generals: list[int],
        cities: list[int],
        graph: ig.Graph,
    ) -> tuple[list[int], list[int]]:
        vs = graph.vs

        # move armies
        for i, player in enumerate(players):
            attacks = player.game.attacks
            while attacks:
                start, end, is50 = attacks.popleft()
                if move_army(graph, i, start, end, is50):
                    break

        # update generals and cities
        if turn % 2 == 1:
            for v in vs[generals]:
                v["army"] += 1
            for v in vs[cities]:
                v["army"] += 1

        # update lands
        if turn > 1 and turn % 50 == 1:
            for v in vs.select(terrain_ge=0):
                v["army"] += 1

        map_new = [col, row]
        map_new += vs["army"]
        map_new += vs["terrain"]

        map_diff = make_diff(map_new, map_old)

        return map_new, map_diff

    # ============================================================
    # run
    # ============================================================

    async def connect(self, client: ClientProtocol, user_id: str) -> None:
        try:
            self.user_ids.put(client, user_id)
        except ValueDuplicationError:
            raise
        except KeyDuplicationError:
            raise
        if user_id not in self.users:
            self.users[user_id] = User(user_id)
        user = self.users[user_id]
        self.players[user_id] = Player(client, user)

    async def disconnect(self, client: ClientProtocol, user_id: str) -> None:
        if self.user_ids[client] != user_id:
            raise
        del self.user_ids[client]

    async def run(self, players: list[Player]) -> None:
        row = self.row
        col = self.col

        graph = make_2d_grid(row, col)
        self.random_connected_map(graph, 0)

        map_: list[int] = []
        generals: list[int] = self.dispersion_generals(2, row, col, graph)
        cities: list[int] = []

        t_start = time.monotonic()
        secs_per_move = self.tps / self.mpt

        self._game_start(players)

        for turn in range(1, 52):
            map_, map_diff = self._map_update(
                players, row, col, turn, map_, generals, cities, graph
            )
            t_start += secs_per_move
            t_end = time.monotonic()
            await asyncio.sleep(max(0, t_start - t_end))
            self._game_update(players, turn, generals, map_diff)

            if __debug__:
                pass

        self._game_over(players)


class SocketioServer:
    def __init__(self) -> None:
        self.sockets: dict[ClientProtocol, AsyncClient] = {}

    # ============================================================
    # send
    # ============================================================

    def set_username(
        self, client: ClientProtocol, user_id: str, username: str
    ) -> asyncio.Task[None]:
        return asyncio.create_task(
            self.sockets[client].emit("set_username", (user_id, username, BOTKEY))
        )

    def stars_and_rank(
        self, client: ClientProtocol, user_id: str
    ) -> asyncio.Task[None]:
        return asyncio.create_task(self.sockets[client].emit("stars_and_rank", user_id))

    def join_private(
        self, client: ClientProtocol, queue_id: str, user_id: str
    ) -> asyncio.Task[None]:
        print(f"https://generals.io/games/{queue_id}")
        return asyncio.create_task(
            self.sockets[client].emit("join_private", (queue_id, user_id, BOTKEY))
        )

    def set_force_start(
        self, client: ClientProtocol, queue_id, do_force: bool
    ) -> asyncio.Task[None]:
        return asyncio.create_task(
            self.sockets[client].emit("set_force_start", (queue_id, do_force))
        )

    def leave_game(self, client: ClientProtocol) -> asyncio.Task[None]:
        return asyncio.create_task(self.sockets[client].emit("leave_game"))

    def surrender(self, client: ClientProtocol) -> asyncio.Task[None]:
        return asyncio.create_task(self.sockets[client].emit("surrender"))

    def attack(
        self, client: ClientProtocol, start: int, end: int, is50: bool
    ) -> asyncio.Task[None]:
        return asyncio.create_task(
            self.sockets[client].emit("attack", (start, end, is50))
        )

    # ============================================================
    # run
    # ============================================================

    async def connect(self, client: ClientProtocol, user_id: str) -> None:
        c = AsyncClient(ssl_verify=False)
        self.sockets[client] = c
        c.event(client.stars)
        c.event(client.rank)
        c.event(client.chat_message)
        c.event(client.notify)
        c.event(client.queue_update)
        c.event(client.pre_game_start)
        c.event(client.game_start)
        c.event(client.game_update)
        c.event(client.game_won)
        c.event(client.game_lost)
        c.event(client.game_over)
        await c.connect(WSURL, transports=["websocket"])

    async def disconnect(self, client: ClientProtocol, user_id: str) -> None:
        c = self.sockets[client]
        del self.sockets[client]
        await c.disconnect()
        await c.wait()

    async def run(self) -> None:
        pass



from asyncio import Event as AEvent
from asyncio import Queue as AQueue
from asyncio import QueueEmpty, Task, create_task
from functools import partial, update_wrapper, wraps
from queue import Empty, Queue
from threading import Event
from typing import Any, Callable, Concatenate, Coroutine, Protocol, Self, overload

__all__ = (
    "methodlike",
    "streamify",
    "signalize",
    "astreamify",
    "asignalize",
    "to_coro",
    "to_task",
)


# ============================================================
# typing
# ============================================================

type _MethodType[T, **P, R] = Callable[Concatenate[T, P], R]


class _MethodLike[T, **P, R](Protocol):
    @overload
    def __get__(
        self,
        instance: None,
        owner: type[T],
        /,
    ) -> _MethodType[T, P, R]: ...

    @overload
    def __get__(
        self,
        instance: T,
        owner: type[T] | None = None,
        /,
    ) -> Callable[P, R]: ...


class _MethodDescriptor[T, **P, R](_MethodLike[T, P, R]):
    def __set_name__(self, owner: type[T], name: str) -> None:
        self.name = "_" + name

    def __init__(self, wrapped: _MethodType[T, P, R]) -> None:
        self.wrapped = wrapped


def methodlike[T, **P, R](m: _MethodType[T, P, R]) -> _MethodLike[T, P, R]:
    return m


# ============================================================
# synchronous
# ============================================================


class _Stream[**P, R]:
    __wrapped__: Callable[P, R]

    def __init__(self, wrapped: Callable[P, R]) -> None:
        update_wrapper(self, wrapped)
        self._queue: Queue[R | None] = Queue()

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        item = self.__wrapped__(*args, **kwargs)
        self._queue.put_nowait(item)
        return item

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> R:
        q = self._queue
        item = q.get()
        q.task_done()
        if item is None:
            raise StopIteration()
        return item

    def wait(self) -> R | None:
        q = self._queue
        item = q.get()
        q.task_done()
        return item

    def get(self) -> R | None:
        q = self._queue
        try:
            item = q.get_nowait()
        except Empty:
            return None
        else:
            q.task_done()
            return item

    def close(self) -> None:
        self._queue.put_nowait(None)


class _Signal[**P, R]:
    __wrapped__: Callable[P, R]

    def __init__(self, wrapped: Callable[P, R]) -> None:
        update_wrapper(self, wrapped)
        self.event = Event()

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        self.event.set()
        return self.__wrapped__(*args, **kwargs)

    def get(self) -> bool:
        s = self.event.is_set()
        if s:
            self.event.clear()
        return s

    def wait(self) -> None:
        self.event.wait()
        self.event.clear()


class streamify[T, **P, R](_MethodDescriptor[T, P, R]):
    @overload
    def __get__(
        self,
        instance: None,
        owner: type[T],
    ) -> _MethodType[T, P, R]: ...

    @overload
    def __get__(
        self,
        instance: T,
        owner: type[T] | None = None,
    ) -> _Stream[P, R]: ...

    def __get__(
        self,
        instance: T | None,
        owner: type[T] | None = None,
    ) -> Any:
        if instance is None:
            return self.wrapped
        try:
            x = getattr(instance, self.name)
        except AttributeError:
            x = _Stream(partial(self.wrapped, instance))
            setattr(instance, self.name, x)
        return x


class signalize[T, **P, R](_MethodDescriptor[T, P, R]):
    @overload
    def __get__(
        self,
        instance: None,
        owner: type[T],
    ) -> _MethodType[T, P, R]: ...

    @overload
    def __get__(
        self,
        instance: T,
        owner: type[T] | None = None,
    ) -> _Signal[P, R]: ...

    def __get__(
        self,
        instance: T | None,
        owner: type[T] | None = None,
    ) -> Any:
        if instance is None:
            return self.wrapped
        try:
            x = getattr(instance, self.name)
        except AttributeError:
            x = _Signal(partial(self.wrapped, instance))
            setattr(instance, self.name, x)
        return x


# ============================================================
# asynchronous
# ============================================================


class _AStream[**P, R]:
    __wrapped__: Callable[P, R]

    def __init__(self, wrapped: Callable[P, R]) -> None:
        update_wrapper(self, wrapped)
        self._queue: AQueue[R | None] = AQueue()

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        item = self.__wrapped__(*args, **kwargs)
        self._queue.put_nowait(item)
        return item

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> R:
        q = self._queue
        item = await q.get()
        q.task_done()
        if item is None:
            raise StopAsyncIteration()
        return item

    async def wait(self) -> R | None:
        q = self._queue
        item = await q.get()
        q.task_done()
        return item

    def get(self) -> R | None:
        q = self._queue
        try:
            item = q.get_nowait()
        except QueueEmpty:
            return None
        else:
            q.task_done()
            return item

    def close(self) -> None:
        self._queue.put_nowait(None)


class _ASignal[**P, R]:
    __wrapped__: Callable[P, R]

    def __init__(self, wrapped: Callable[P, R]) -> None:
        update_wrapper(self, wrapped)
        self.event = AEvent()

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        self.event.set()
        return self.__wrapped__(*args, **kwargs)

    def get(self) -> bool:
        s = self.event.is_set()
        if s:
            self.event.clear()
        return s

    async def wait(self) -> None:
        await self.event.wait()
        self.event.clear()


class astreamify[T, **P, R](_MethodDescriptor[T, P, R]):
    @overload
    def __get__(
        self,
        instance: None,
        owner: type[T],
    ) -> _MethodType[T, P, R]: ...

    @overload
    def __get__(
        self,
        instance: T,
        owner: type[T] | None = None,
    ) -> _AStream[P, R]: ...

    def __get__(
        self,
        instance: T | None,
        owner: type[T] | None = None,
    ) -> Any:
        if instance is None:
            return self.wrapped
        try:
            x = getattr(instance, self.name)
        except AttributeError:
            x = _AStream(partial(self.wrapped, instance))
            setattr(instance, self.name, x)
        return x


class asignalize[T, **P, R](_MethodDescriptor[T, P, R]):
    @overload
    def __get__(
        self,
        instance: None,
        owner: type[T],
    ) -> _MethodType[T, P, R]: ...

    @overload
    def __get__(
        self,
        instance: T,
        owner: type[T] | None = None,
    ) -> _ASignal[P, R]: ...

    def __get__(
        self,
        instance: T | None,
        owner: type[T] | None = None,
    ) -> Any:
        if instance is None:
            return self.wrapped
        try:
            x = getattr(instance, self.name)
        except AttributeError:
            x = _ASignal(partial(self.wrapped, instance))
            setattr(instance, self.name, x)
        return x


def to_coro[**P, R](wrapped: Callable[P, R]) -> Callable[P, Coroutine[Any, Any, R]]:
    @wraps(wrapped)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        return wrapped(*args, **kwargs)

    return wrapper


def to_task[**P, R](
    wrapped: Callable[P, Coroutine[Any, Any, R]],
) -> Callable[P, Task[R]]:
    @wraps(wrapped)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Task[R]:
        return create_task(wrapped(*args, **kwargs))

    return wrapper



import asyncio
from typing import Any

from socketio import AsyncClient

from glhf.base import BotProtocol, ClientProtocol, ServerProtocol
from glhf.gui import PygameGUI
from glhf.typing_ import GameStartDict, GameUpdateDict, QueueUpdateDict

WSURL = "wss://ws.generals.io"
BOTKEY = "sd09fjd203i0ejwi_changeme"


class BasicClient(ClientProtocol):
    def __init__(
        self,
        userid: str,
        username: str,
        bot: BotProtocol,
        gui: PygameGUI,
        server: ServerProtocol,
    ) -> None:
        self.userid = userid
        self.username = username
        self.queue_id = ""
        self.server = server
        self.bot = bot
        self.gui = gui

    # ============================================================
    # recieve
    # ============================================================

    def stars(self, data: dict[str, float]) -> None:
        self.bot.stars(data)

    def rank(self, data: dict[str, int]) -> None:
        self.bot.rank(data)

    def chat_message(self, chat_room: str, data: dict[str, Any]) -> None:
        self.bot.chat_message(chat_room, data)

    def notify(self, data: Any, _: Any = None) -> None:
        self.bot.notify(data)

    def queue_update(self, data: QueueUpdateDict) -> None:
        self.bot.queue_update(data)

    def pre_game_start(self) -> None:
        self.bot.pre_game_start()

    def game_start(self, data: GameStartDict, _: Any = None) -> None:
        self.bot.game_start(data)
        self.gui.game_start(data)

    def game_update(self, data: GameUpdateDict, _: Any = None) -> None:
        self.bot.game_update(data)
        self.gui.game_update(data)

    def game_won(self, _1: Any = None, _2: Any = None) -> None:
        self.bot.game_won()

    def game_lost(self, _1: Any = None, _2: Any = None) -> None:
        self.bot.game_lost()

    def game_over(self, _1: Any = None, _2: Any = None) -> None:
        self.bot.game_over()

    # ============================================================
    # send
    # ============================================================

    def set_username(self) -> asyncio.Task[None]:
        return self.server.set_username(self, self.userid, self.username)

    def stars_and_rank(self) -> asyncio.Task[None]:
        return self.server.stars_and_rank(self, self.userid)

    def join_private(self, queue_id: str) -> asyncio.Task[None]:
        self.queue_id = queue_id
        return self.server.join_private(self, queue_id, self.userid)

    def set_force_start(self, do_force: bool) -> asyncio.Task[None]:
        assert self.queue_id
        return self.server.set_force_start(self, self.queue_id, do_force)

    def leave_game(self) -> asyncio.Task[None]:
        return self.server.leave_game(self)

    def surrender(self) -> asyncio.Task[None]:
        return self.server.surrender(self)

    def attack(self, start: int, end: int, is50: bool) -> asyncio.Task[None]:
        return self.server.attack(self, start, end, is50)

    # ============================================================
    # run
    # ============================================================

    def __hash__(self) -> int:
        return hash(self.userid)

    async def __aenter__(self) -> None:
        self.gui.__enter__()
        await self.server.connect(self, self.userid)

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self.gui.__exit__(exc_type, exc_val, exc_tb)
        await self.server.disconnect(self, self.userid)

    async def run(self) -> None:
        async with self:
            await self.bot.run(self)


class SocketioClient(AsyncClient, ClientProtocol):
    def __init__(
        self,
        userid: str,
        username: str,
        bot: BotProtocol,
        gui: PygameGUI,
        /,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs, ssl_verify=False)

        self.userid = userid
        self.username = username
        self.queue_id = ""
        self.bot = bot
        self.gui = gui

        self.event(self.stars)
        self.event(self.rank)
        self.event(self.chat_message)
        self.event(self.notify)
        self.event(self.queue_update)
        self.event(self.pre_game_start)
        self.event(self.game_start)
        self.event(self.game_update)
        self.event(self.game_won)
        self.event(self.game_lost)
        self.event(self.game_over)

    # ============================================================
    # recieve
    # ============================================================

    def stars(self, data: dict[str, float]) -> None:
        self.bot.stars(data)

    def rank(self, data: dict[str, int]) -> None:
        self.bot.rank(data)

    def chat_message(self, chat_room: str, data: dict[str, Any]) -> None:
        self.bot.chat_message(chat_room, data)

    def notify(self, data: Any, _: Any = None) -> None:
        self.bot.notify(data)

    def queue_update(self, data: QueueUpdateDict) -> None:
        self.bot.queue_update(data)

    def pre_game_start(self) -> None:
        self.bot.pre_game_start()

    def game_start(self, data: GameStartDict, _: Any = None) -> None:
        self.bot.game_start(data)
        self.gui.game_start(data)

    def game_update(self, data: GameUpdateDict, _: Any = None) -> None:
        self.bot.game_update(data)
        self.gui.game_update(data)

    def game_won(self, _1: Any = None, _2: Any = None) -> None:
        self.bot.game_won()

    def game_lost(self, _1: Any = None, _2: Any = None) -> None:
        self.bot.game_lost()

    def game_over(self, _1: Any = None, _2: Any = None) -> None:
        self.bot.game_over()

    # ============================================================
    # emit
    # ============================================================

    def set_username(self) -> asyncio.Task[None]:
        return asyncio.create_task(
            self.emit("set_username", (self.userid, self.username, BOTKEY))
        )

    def stars_and_rank(self) -> asyncio.Task[None]:
        return asyncio.create_task(self.emit("stars_and_rank", self.userid))

    def join_private(self, queue_id: str) -> asyncio.Task[None]:
        self.queue_id = queue_id
        print(f"https://generals.io/games/{queue_id}")
        return asyncio.create_task(
            self.emit("join_private", (queue_id, self.userid, BOTKEY))
        )

    def set_force_start(self, do_force: bool) -> asyncio.Task[None]:
        assert self.queue_id
        return asyncio.create_task(
            self.emit("set_force_start", (self.queue_id, do_force))
        )

    def leave_game(self) -> asyncio.Task[None]:
        return asyncio.create_task(self.emit("leave_game"))

    def surrender(self) -> asyncio.Task[None]:
        return asyncio.create_task(self.emit("surrender"))

    def attack(self, start: int, end: int, is50: bool) -> asyncio.Task[None]:
        return asyncio.create_task(self.emit("attack", (start, end, is50)))

    # ============================================================
    # run
    # ============================================================

    async def __aenter__(self) -> None:
        self.gui.__enter__()
        await self.connect(WSURL, transports=["websocket"])

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self.gui.__exit__(exc_type, exc_val, exc_tb)
        await self.disconnect()
        await self.wait()

    async def run(self) -> None:
        async with self:
            await self.bot.run(self)



from typing import Any

from glhf.base import BotProtocol
from glhf.typing_ import GameStartDict, GameUpdateDict, QueueUpdateDict
from glhf.utils import asignalize, astreamify


class Bot(BotProtocol):
    """The `Bot` class is a subclass of `BotProtocol`. It allows customization by overriding the `run` method for specific game interactions. Alternatively, you can create a class that adheres to the `BotProtocol` interface.

    Subclassing `Bot` provides additional functionality:

    The `queue_update`, `game_start`, and `game_update` methods are enhanced with the `astreamify` decorator. This decorator transforms these methods into an async generator, which can be used in an async for loop for server update processing.

    Example:
        ```python
        class MyBot(Bot):
            async def run(self, client: ClientProtocol) -> None:
                ...
                # process queue updates
                async for data in self.queue_update:
                    if not data["isForcing"]:
                        client.set_force_start(True)

                # wait for game start
                await self.game_start.wait()

                # process game updates
                map_ = []
                cities = []
                async for data in self.game_update:
                    map_ = patch(map_, data["map_diff"])
                    cities = patch(cities, data["cities_diff"])
                ...
        ```

    See Also:
        - `BotProtocol`
        - `astreamify`
        - `asignalize`

    """

    # ============================================================
    # recieve
    # ============================================================

    def stars(self, data: dict[str, float]) -> None:
        pass

    def rank(self, data: dict[str, int]) -> None:
        pass

    def chat_message(self, chat_room: str, data: dict[str, Any]) -> None:
        pass

    def notify(self, data: Any) -> None:
        pass

    @astreamify
    def queue_update(self, data: QueueUpdateDict) -> QueueUpdateDict:
        return data

    def pre_game_start(self) -> None:
        pass

    @astreamify
    def game_start(self, data: GameStartDict) -> GameStartDict:
        self.queue_update.close()
        return data

    @astreamify
    def game_update(self, data: GameUpdateDict) -> GameUpdateDict:
        return data

    @asignalize
    def game_won(self) -> None:
        pass

    @asignalize
    def game_lost(self) -> None:
        pass

    @asignalize
    def game_over(self) -> None:
        self.game_update.close()
